﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Media.Animation;
using System.Windows.Media.Effects;

namespace BlogPostTwistori
{
  public partial class ExplodingTextControl : UserControl
  {
    public event EventHandler Exploded;
    public event EventHandler Imploded;

    public ExplodingTextControl()
    {
      InitializeComponent();

      transformFactors = new TransformFactors();
    }
    public static DependencyProperty TextProperty = DependencyProperty.Register(
      "Text", typeof(string), typeof(ExplodingTextControl),
      new PropertyMetadata(OnTextChanged));

    public static DependencyProperty IsInitiallyExplodedProperty = DependencyProperty.Register(
      "IsInitiallyExploded", typeof(bool), typeof(ExplodingTextControl),
      new PropertyMetadata(true, OnTextChanged));

    public static DependencyProperty DurationProperty = DependencyProperty.Register(
      "Duration", typeof(Duration), typeof(ExplodingTextControl),
      new PropertyMetadata(new Duration(new TimeSpan(0, 0, 5)), OnDeferredPropertyChanged));

    public string Text
    {
      get
      {
        return ((string)base.GetValue(ExplodingTextControl.TextProperty));
      }
      set
      {
        base.SetValue(ExplodingTextControl.TextProperty, value);
      }
    }
    public bool IsInitiallyExploded 
    {
      get
      {
        return ((bool)base.GetValue(ExplodingTextControl.IsInitiallyExplodedProperty));
      }
      set
      {
        base.SetValue(ExplodingTextControl.IsInitiallyExplodedProperty, value);
      }
    }
    public Duration Duration
    {
      get
      {
        return ((Duration)base.GetValue(ExplodingTextControl.DurationProperty));
      }
      set
      {
        base.SetValue(ExplodingTextControl.DurationProperty, value);
      }
    } 
    public void Implode()
    {
      StopStoryboards();
      RecreateStoryboardsIfDirty();
      implodeStoryboard.Begin();
    }
    public void Explode()
    {
      StopStoryboards();
      RecreateStoryboardsIfDirty();
      explodeStoryboard.Begin();
    }
    void RecreateStoryboardsIfDirty()
    {
      if (deferredPropertyChanged)
      {
        CreateStoryboardsFromText();
      }
    }
    private void StopStoryboards()
    {
      if (explodeStoryboard != null)
      {
        explodeStoryboard.Stop();
      }
      if (implodeStoryboard != null)
      {
        implodeStoryboard.Stop();
      }
    }
    void CreateStoryboardsFromText()
    {
      txtText.Inlines.Clear();
      CreateImplodeExplodeStoryboards();

      if (!string.IsNullOrEmpty(Text))
      {
        string[] words = Text.Split(' ');

        for (int i = 0; i < words.Length; i++)
        {
          // Was going to use Run here but can't find RenderTransformOrigin for that.
          TextBlock textBlock = new TextBlock()
          {
            RenderTransformOrigin = new Point(0.5, 0.5),
            Text = words[i] + ((i < words.Length - 1) ? " " : string.Empty)
          };
          txtText.Inlines.Add(textBlock);

          AddTransformationAndAnimationToTextBlock(textBlock);          
        }
      }
      deferredPropertyChanged = false;
    }
    void CreateImplodeExplodeStoryboards()
    {
      implodeStoryboard = CreateNewStoryboard();
      explodeStoryboard = CreateNewStoryboard();

      implodeStoryboard.Completed += (s, e) =>
        {
          if (Imploded != null)
          {
            Imploded(this, null);
          }
        };

      explodeStoryboard.Completed += (s, e) =>
        {
          if (Exploded != null)
          {
            Exploded(this, null);
          }
        };
    }
    private void AddTransformationAndAnimationToTextBlock(TextBlock textBlock)
    {
      transformFactors.Randomise();

      AddTransformationsToTextBlock(textBlock);

      AddAnimationsForTextBlockToStoryboard(textBlock);
    }
    private void AddAnimationsForTextBlockToStoryboard(TextBlock textBlock)
    {
      // TODO: Might be nicer to not use strings here.
      AddStoryboardTextBlockAnimation(textBlock,
        "RenderTransform.Children[0].ScaleX", transformFactors.Scale, 1.0);

      AddStoryboardTextBlockAnimation(textBlock, 
        "RenderTransform.Children[0].ScaleY", transformFactors.Scale, 1.0);

      AddStoryboardTextBlockAnimation(textBlock,
        "RenderTransform.Children[1].X", transformFactors.TranslateX, 0.0);

      AddStoryboardTextBlockAnimation(textBlock,
        "RenderTransform.Children[1].Y", transformFactors.TranslateY, 0.0);

      AddStoryboardTextBlockAnimation(textBlock,
        "RenderTransform.Children[2].Angle", transformFactors.Rotation, 0.0);

      AddStoryboardTextBlockAnimation(textBlock, "Opacity", 0.0, 1.0);
    }
    void AddStoryboardTextBlockAnimation(TextBlock textEffect,
      string property, double implodeFrom, double implodeTo)
    {
      DoubleAnimation animation = CreateDoubleAnimationForTextBlock(textEffect, property,
        implodeFrom, implodeTo);

      animation.Duration = this.Duration; // TODO: Why?

      implodeStoryboard.Children.Add(animation);

      animation = CreateDoubleAnimationForTextBlock(textEffect, property,
        implodeTo, implodeFrom);

      animation.Duration = this.Duration; // TODO: Why?

      explodeStoryboard.Children.Add(animation);
    }
    static DoubleAnimation CreateDoubleAnimationForTextBlock(TextBlock textEffect, string property,
      double from, double to)
    {
      DoubleAnimation animation = new DoubleAnimation();
      animation.From = from;
      animation.To = to;
      Storyboard.SetTarget(animation, textEffect);
      Storyboard.SetTargetProperty(animation, new PropertyPath(property));
      return (animation);
    }
    void AddTransformationsToTextBlock(TextBlock run)
    {
      ScaleTransform scale = new ScaleTransform() { CenterX = 0.5, CenterY = 0.5 };
      TranslateTransform translate = new TranslateTransform();
      RotateTransform rotate = new RotateTransform();

      if (IsInitiallyExploded)
      {
        transformFactors.ApplyTo(scale);
        transformFactors.ApplyTo(translate);
        transformFactors.ApplyTo(rotate);        
      }

      run.RenderTransform = new TransformGroup()
      {
        Children = new TransformCollection() 
        {
          scale,
          translate,
          rotate
        }        
      };
    }
    private Storyboard CreateNewStoryboard()
    {
      Storyboard sb = new Storyboard()
      {
        FillBehavior = FillBehavior.HoldEnd,
        BeginTime = new TimeSpan(0),
        SlipBehavior = SlipBehavior.Slip
      };
      sb.SetValue(Storyboard.DurationProperty, this.Duration);
      return (sb);
    }
    static void OnTextChanged(DependencyObject sender,
      DependencyPropertyChangedEventArgs args)
    {
      ExplodingTextControl control = (ExplodingTextControl)sender;
      control.CreateStoryboardsFromText();
    }
    static void OnDeferredPropertyChanged(DependencyObject sender, 
      DependencyPropertyChangedEventArgs args)
    {
      ExplodingTextControl control = (ExplodingTextControl)sender;
      control.deferredPropertyChanged = true;
    }
    bool deferredPropertyChanged;
    Storyboard implodeStoryboard;
    Storyboard explodeStoryboard;
    TransformFactors transformFactors;
    const double MAX_SHADOW_DEPTH = 20.0;
  }
}